#include "list.h"

#include <assert.h>
#include <string.h>

void list_init(struct list *list, size_t element_size)
{
    assert(list != NULL);
    assert(element_size > 0);

    list->head = list->tail = NULL;
    list->elem_size = element_size;
}

void list_destroy(struct list *list)
{
    assert(list != NULL);

    // Let ‹list_pop_front()› take care of node removal.
    while (list_pop_front(list, NULL))
        /* nop */;
}

size_t list_size(const struct list *list)
{
    assert(list != NULL);

    size_t counter = 0U;

    for (const struct node *node = list->head; node != NULL; node = node->next)
        ++counter;

    return counter;
}

bool list_is_empty(const struct list *list)
{
    assert(list != NULL);
    return list->head == NULL;
}

static
struct node *list_allocate_node(const struct list *list, const void *data)
{
    struct node *node = malloc(sizeof(struct node));

    if (node == NULL)
        return NULL;

    node->data = malloc(list->elem_size);

    if (node->data == NULL) {
        free(node);
        return NULL;
    }

    memcpy(node->data, data, list->elem_size);
    return node;
}

static
void list_release_node(const struct list *list, struct node *node, void *data)
{
    if (data != NULL)
        memcpy(data, node->data, list->elem_size);

    free(node->data);
    free(node);
}

bool list_push_back(struct list *list, const void *data)
{
    assert(list != NULL);
    assert(data != NULL);

    struct node *node = list_allocate_node(list, data);

    if (node == NULL)
        return false;

    node->next = NULL;

    // Attach the node after ‹list->tail›.
    if (list->tail != NULL)
        list->tail->next = node;
    else
        list->head = node;

    list->tail = node;
    return true;
}

bool list_pop_back(struct list *list, void *data)
{
    assert(list != NULL);

    if (list_is_empty(list))
        return false;

    // We need to traverse the entire list until we find the node
    // that precedes the ‹list->tail›.
    struct node *prev = list->head;
    while (prev != NULL && prev->next != list->tail)
        prev = prev->next;

    // Copy data and remove the current tail node.
    list_release_node(list, list->tail, data);

    // Restore ‹list›'s correctness.
    list->tail = prev;

    if (prev == NULL)
        list->head = NULL;
    else
        prev->next = NULL;

    return true;
}

bool list_push_front(struct list *list, const void *data)
{
    assert(list != NULL);
    assert(data != NULL);

    // This function works analogously to ‹list_push_back()›.
    struct node *node = list_allocate_node(list, data);

    if (node == NULL)
        return false;

    // Link the new ‹node› into the list.
    node->next = list->head;

    if (list->head == NULL)
        list->tail = node;

    list->head = node;
    return true;
}

bool list_pop_front(struct list *list, void *data)
{
    assert(list != NULL);

    if (list_is_empty(list))
        return false;

    // Remember node after the current head.
    struct node *next = list->head->next;

    // Copy data and remove the current head node.
    list_release_node(list, list->head, data);

    // Restore ‹list›'s correctness.
    list->head = next;

    if (list->head == NULL)
        list->tail = NULL;

    return true;
}
